#include "LogNormalPoissonDistribution.h"
#include <cmath>
#include <boost/math/distributions/normal.hpp>

using namespace std;

const char* const LogNormalPoissonDistribution::valparnams[]={"mu","sigma","VaR"};

void LogNormalPoissonDistribution::setfixed(unsigned int i,long double val){
  switch(i){
  case 0:
    mu=val;
    fixedmu=1;
    break;
  case 1:
    sigma=val;
    fixedsig=1;
    break;
  default:
    throw Distribution_error("Fixing VaR not currently possible for normal distribution.\n");
  };
};

void LogNormalPoissonDistribution::unfix(unsigned int i){
  switch(i){
  case 0:
    fixedmu=0;
    break;
  case 1:
    fixedsig=0;
    break;
  default:
    break;
    //VaR can't be fixed anyway.
    //Perhaps should give some indication of a problem.
  };
};

void LogNormalPoissonDistribution::setmain(unsigned int i,long double at){
  mainpar=i;
  if(i==2){
    VaR_at=at;
  };
};



long double LogNormalPoissonDistribution::parval() const{
  switch(mainpar){
  case 0:
    return mu;
  case 1:
    return sigma;
  default: //could check for corruption.
    return this->quant(VaR_at);
  };
};

long double LogNormalPoissonDistribution::parval(unsigned int i,long double at) const{
  switch(i){
  case 0:
    return mu;
  case 1:
    return sigma;
  default:
    return this->quant(at);
  };
};

LogNormalPoissonDistribution* LogNormalPoissonDistribution::copy() const{
  LogNormalPoissonDistribution *answer=new LogNormalPoissonDistribution(mu,sigma);
  return answer;
};



long double *LogNormalPoissonDistribution::dist(const Data &dat) const{
  const unsigned int *data=(const unsigned int *)dat.getvalues();
  unsigned int x=0;
  for(unsigned int i=0;i<dat.n;i++){
    if(*(data+i)>x){
      x=*(data+i);
    };
  };
  long double *probs=new long double[x+1];
  this->evalprobstodirect(x,probs);
  //calculates P(X=n) for each n
  long double *cumprobs=new long double[x+1];
  *cumprobs=*probs;
  for(unsigned int i=1;i<=x;i++){
    *(cumprobs+i)=*(cumprobs+i-1)+*(probs+i);
  };
  long double *ans=new long double[dat.n];
  for(unsigned int i=0;i<dat.n;i++){
    *(ans+i)=*(cumprobs+*(data+i));
  };
  delete[] probs;
  delete[] cumprobs;
  return ans;
};


long double *LogNormalPoissonDistribution::survive(const Data &dat) const{
  long double *ans=this->dist(dat);
  for(unsigned int i=0;i<dat.n;i++){
    *(ans+i)=1-*(ans+i);
  };
  return ans;
};

long double LogNormalPoissonDistribution::quant(long double p) const{
  unsigned int ans=0;
  //Need to figure out how to do this.
  return ans;  
};


//Need some helper functions for estimating MLE.

void LogNormalPoissonDistribution::evalprobstodirect(unsigned int x,long double *answers) const{
  //calculates the probability of all observations up to x.
  //answers should be a vector of length x+1
  //integrates out over log(lambda)
  static long double tol=log(2)-log(LDBL_EPSILON);
  long double upper=sqrt(2*x+4*tol);//Integral beyond this point should be negligible. 
  long double lower=-upper;//Integral before this point should be negligible.
  //Calculate coefficients

  for(unsigned int i=0;i<=x;i++){
    *(answers+i)=0;
  };
    
  long double lambda=exp(mu);


  unsigned int numpoints=1000;
  long double stepsize=(upper-lower)/(numpoints-1);
  long double r=lower;

  for(unsigned int i=0;i<numpoints;i++){
    long double lda=lambda*exp(r*sigma);
    long double lexrs=lambda*exp(r*sigma)+r*r/2;
    long double exex=exp(-lexrs);      
    for(unsigned int j=0;j<=x;j++){
      *(answers+j)+=exex;
      exex*=lda/(j+1);
    };
    r+=stepsize;
  };
  for(unsigned int j=0;j<=x;j++){
    *(answers+j)*=stepsize/sqrt(2*3.141592653589793238462643383);
  };
};


void LogNormalPoissonDistribution::evalprobsandderivstodirect(unsigned int x,long double *answers) const{
  //calculates the probability of all observations up to x.
  //answers should be a vector of length x+1
  //integrates out over log(lambda)
  static long double tol=log(2)-log(LDBL_EPSILON);
  long double upper=sqrt(2*x+4*tol);//Integral beyond this point should be negligible. 
  long double lower=-upper;//Integral before this point should be negligible.
  //Calculate coefficients

  for(unsigned int i=0;i<(x+1)*6;i++){
    *(answers+i)=0;
  };
    
  long double lambda=exp(mu);


  unsigned int numpoints=1000;
  long double stepsize=(upper-lower)/(numpoints-1);
  long double r=lower;

  for(unsigned int i=0;i<numpoints;i++){
    long double lda=lambda*exp(r*sigma);
    long double lexrs=lambda*exp(r*sigma)+r*r/2;
    long double exex=exp(-lexrs);      
    for(unsigned int j=0;j<=x;j++){
    long double xminlda=j-lda;
      *(answers+j*6)+=exex;
      *(answers+j*6+1)+=exex*xminlda;
      *(answers+j*6+2)+=exex*r*xminlda;
      *(answers+j*6+3)+=exex*(xminlda*xminlda-lda);
      *(answers+j*6+4)+=exex*r*(xminlda*xminlda-lda);
      *(answers+j*6+5)+=exex*r*r*(xminlda*xminlda-lda);
      exex*=lda/(j+1);
    };
    r+=stepsize;
  };
  for(unsigned int j=0;j<(x+1)*6;j++){
    *(answers+j)*=stepsize/sqrt(2*3.141592653589793238462643383);
  };
};


void LogNormalPoissonDistribution::evalprobsandderivstodirectsparse(unsigned int x,const unsigned int *values,long double *answers)const{
  //calculates the probability of all observations in vector values.
  //answers should be a vector of length x.
  //integrates out over log(lambda)
  //If the values are very sparse, it may be possible to use a more
  //efficient method.
  static long double tol=log(2)-log(LDBL_EPSILON);
  long double upper=sqrt(2*x+4*tol);//Integral beyond this point should be negligible. 
  long double lower=-upper;//Integral before this point should be negligible.
  //Calculate coefficients

  for(unsigned int i=0;i<(x+1)*6;i++){
    *(answers+i)=0;
  };
    
  long double lambda=exp(mu);


  unsigned int numpoints=1000;
  long double stepsize=(upper-lower)/(numpoints-1);
  long double r=lower;

  for(unsigned int i=0;i<numpoints;i++){
    long double lda=lambda*exp(r*sigma);
    long double lexrs=lambda*exp(r*sigma)+r*r/2;
    long double exex=exp(-lexrs);
    unsigned int dt=0;
    for(unsigned int j=0;j<x;j++){
      for(;dt<*(values+j);dt++){
	exex*=lda/(dt+1);
      };
      long double xminlda=dt-lda;
      *(answers+j*6)+=exex;
      *(answers+j*6+1)+=exex*xminlda;
      *(answers+j*6+2)+=exex*r*xminlda;
      *(answers+j*6+3)+=exex*(xminlda*xminlda-lda);
      *(answers+j*6+4)+=exex*r*(xminlda*xminlda-lda);
      *(answers+j*6+5)+=exex*r*r*(xminlda*xminlda-lda);
    };
    r+=stepsize;
  };
  for(unsigned int j=0;j<(x+1)*6;j++){
    *(answers+j)*=stepsize/sqrt(2*3.141592653589793238462643383);
  };
};



void LogNormalPoissonDistribution::estimate(const Data &dat){
  //Assumes discrete data - program should check for this before calling.  
  const unsigned int *data=(const unsigned int*)dat.getvalues();
  unsigned int maxdata=0;
  unsigned int n=dat.n;
  for(unsigned int i=0;i<n;i++){
    if(*(data+i)>maxdata){
      maxdata=*(data+i);
    };
  };
  if(dat.get_type().storage&(LIST|FREQUENCY_TABLE)){
    unsigned int *freqs=new unsigned int[maxdata+1];
    for(unsigned int i=0;i<=maxdata;i++){
      *(freqs+i)=0;
    };
    const unsigned int *frqs=dat.getfreqs();
    if(frqs){
      n=0;
      for(unsigned int i=0;i<dat.n;i++){
	(*(freqs+*(data+i)))+=*(frqs+i);
	n+=*(frqs+i);
      };
    }else{
      for(unsigned int i=0;i<dat.n;i++){
	(*(freqs+*(data+i)))++;
      };
    };
    long double totlog=-(long double)(*freqs);//Arbitrarily set log(0)=-1.
    long double totlogsq=(*freqs);//Can probably do better ??
    for(unsigned int i=1;i<=maxdata;i++){
      if(*(freqs+i)){
	totlog+=log(i)*(*(freqs+i));
	totlogsq+=log(i)*log(i)*(*(freqs+i));
      };
    };
      
    mu=totlog/n;
    long double sigsq=totlogsq/n-mu*mu;
    sigma=sqrt(sigsq);
      
    long double*probsandderivs=new long double[(maxdata+1)*6];
    this->evalprobsandderivstodirect(maxdata,probsandderivs);
      
    const unsigned int maxiter=10000;
    long double dmul=0;      
    long double dsigl=0;	   
    long double ddmul=0;	   
    long double ddsigl=0;	   
    long double ddmusigl=0;  
    long double xdmul=0;
    long double xdsigl=0;
    long double xddmul=0;
    long double xddsigl=0; 
    long double xddmusigl=0;
    for(unsigned int i=0;i<=maxdata;i++){
      if(*(freqs+i)){
	
	
	xdmul=*(probsandderivs+6*i+1)/(*(probsandderivs+6*i));
	xdsigl=*(probsandderivs+6*i+2)/(*(probsandderivs+6*i));
	xddmul=*(probsandderivs+6*i+3)/(*(probsandderivs+6*i))-*(probsandderivs+6*i+1)*(*(probsandderivs+6*i+1))/(*(probsandderivs+6*i)*(*(probsandderivs+6*i)));
	xddmusigl=*(probsandderivs+6*i+4)-*(probsandderivs+6*i+1)*(*(probsandderivs+6*i+2))/(*(probsandderivs+6*i)*(*(probsandderivs+6*i)));
	xddsigl=*(probsandderivs+6*i+5)/(*(probsandderivs+6*i))-*(probsandderivs+6*i+2)*(*(probsandderivs+6*i+2))/(*(probsandderivs+6*i)*(*(probsandderivs+6*i)));
      
	dmul+=xdmul*(*(freqs+i));      
	dsigl+=xdsigl*(*(freqs+i));	   	   
	ddmul+=xddmul*(*(freqs+i));	   	   
	ddsigl+=xddsigl*(*(freqs+i));	   	   
	ddmusigl+=xddmusigl*(*(freqs+i));
      };
    };
    for(unsigned int numiter=0;((dmul*dmul>1e-10)||(dsigl*dsigl>1e-10));numiter++){
      if(numiter>=maxiter){
	//Failed to converge
	throw Distribution_error("Could not find MLE estimate for log-normal parameters.\n");
      };
      long double dddet=ddmul*ddsigl-ddmusigl*ddmusigl;
      long double ddtr=ddmul+ddsigl;
      //Check for negative definiteness
      if(dddet>0){
	if(ddtr<0){
	  //negative definite. Should be OK
	  mu-=(ddsigl*dmul-ddmusigl*dsigl)/dddet;
	  long double sigstep=(ddmul*dsigl-ddmusigl*dmul)/dddet;
	  if(sigma<sigstep){
	    sigma/=2;
	  }else{
	    sigma-=sigstep;
	  };
	}else{
	  //positive definite.
	  mu=(ddsigl*dmul-ddmusigl*dsigl)/dddet;
	  long double sigstep=(ddmul*dsigl-ddmusigl*dmul)/dddet;
	  if(sigma+sigstep<0){
	    sigma/=2;
	  }else{
	    sigma+=sigstep;
	  };	
	};
      }else{
	//Saddle point
	//use gradient ascent
	mu+=dmul/abs(ddmul);
	long double sigmastep=dsigl/abs(ddsigl);
	if(sigma+sigmastep>0){
	  sigma+=sigmastep;
	}else{
	  sigma/=2;
	};
      };
      this->evalprobsandderivstodirect(maxdata,probsandderivs);
      dmul=0;      
      dsigl=0;	   
      ddmul=0;	   
      ddsigl=0;	   
      ddmusigl=0;
      for(unsigned int i=0;i<=maxdata;i++){
	if(*(freqs+i)){	    
	  xdmul=*(probsandderivs+6*i+1)/(*(probsandderivs+6*i));
	  xdsigl=*(probsandderivs+6*i+2)/(*(probsandderivs+6*i));
	  xddmul=*(probsandderivs+6*i+3)/(*(probsandderivs+6*i))-*(probsandderivs+6*i+1)*(*(probsandderivs+6*i+1))/(*(probsandderivs+6*i)*(*(probsandderivs+6*i)));
	  xddmusigl=*(probsandderivs+6*i+4)/(*(probsandderivs+6*i))-*(probsandderivs+6*i+1)*(*(probsandderivs+6*i+2))/(*(probsandderivs+6*i)*(*(probsandderivs+6*i)));
	  xddsigl=*(probsandderivs+6*i+5)/(*(probsandderivs+6*i))-*(probsandderivs+6*i+2)*(*(probsandderivs+6*i+2))/(*(probsandderivs+6*i)*(*(probsandderivs+6*i)));
	  dmul+=xdmul*(*(freqs+i));      
	  dsigl+=xdsigl*(*(freqs+i));	   	   
	  ddmul+=xddmul*(*(freqs+i));	   	   
	  ddsigl+=xddsigl*(*(freqs+i));	   	   
	  ddmusigl+=xddmusigl*(*(freqs+i));  	       		
	};
      };
    };
    delete[] freqs;
    delete[] probsandderivs;
  }else if(dat.get_type().storage&GROUPED){
    const unsigned int *freqs=dat.getfreqs();

    //now data is the array of cut-offs and freqs the frequencies.

    long double totlog=(long double)(*freqs)*log((*data)/2);//Arbitrarily set log(0)=-1.
    long double totlogsq=(*freqs)*log((*data)/2)*log((*data)/2);//Can probably do better ??
    for(unsigned int i=1;i<dat.n;i++){
      if(*(freqs+i)){
	totlog+=(*(freqs+i))*log((*(data+i)+*(data+i-1))/2);
	totlogsq+=(*(freqs+i))*log((*(data+i)+*(data+i-1))/2)*log((*(data+i)+*(data+i+1))/2);
      };
    };
    mu=totlog/n;
    long double sigsq=totlogsq/n-mu*mu;
    sigma=sqrt(sigsq);
      
    long double *probsandderivs=new long double[(maxdata+1)*6];
    long double *panddgrouped=new long double[dat.n*6];
    this->evalprobsandderivstodirect(maxdata,probsandderivs);
    unsigned int firstval=0;
    for(unsigned int i=0;i<dat.n;i++){
      for(unsigned int j=0;j<6;j++){
	*(panddgrouped+i*6+j)=0;
      };
      for(unsigned int k=firstval;k<*(data+i);k++){
	for(unsigned int j=0;j<6;j++){
	  *(panddgrouped+i*6+j)+=*(probsandderivs+k*6+j);
	};
      };
    };
    
    const unsigned int maxiter=10000;
    long double dmul=0;      
    long double dsigl=0;	   
    long double ddmul=0;	   
    long double ddsigl=0;	   
    long double ddmusigl=0;  
    long double xdmul=0;
    long double xdsigl=0;
    long double xddmul=0;
    long double xddsigl=0; 
    long double xddmusigl=0;
    for(unsigned int i=0;i<=maxdata;i++){
      if(*(freqs+i)){
	
	
	xdmul=*(panddgrouped+6*i+1)/(*(panddgrouped+6*i));
	xdsigl=*(panddgrouped+6*i+2)/(*(panddgrouped+6*i));
	xddmul=*(panddgrouped+6*i+3)/(*(panddgrouped+6*i))-*(panddgrouped+6*i+1)*(*(panddgrouped+6*i+1))/(*(panddgrouped+6*i)*(*(panddgrouped+6*i)));
	xddmusigl=*(panddgrouped+6*i+4)-*(panddgrouped+6*i+1)*(*(panddgrouped+6*i+2))/(*(panddgrouped+6*i)*(*(panddgrouped+6*i)));
	xddsigl=*(panddgrouped+6*i+5)/(*(panddgrouped+6*i))-*(panddgrouped+6*i+2)*(*(panddgrouped+6*i+2))/(*(panddgrouped+6*i)*(*(panddgrouped+6*i)));
      
	dmul+=xdmul*(*(freqs+i));      
	dsigl+=xdsigl*(*(freqs+i));	   	   
	ddmul+=xddmul*(*(freqs+i));	   	   
	ddsigl+=xddsigl*(*(freqs+i));	   	   
	ddmusigl+=xddmusigl*(*(freqs+i));
      };
    };
    for(unsigned int numiter=0;((dmul*dmul>1e-10)||(dsigl*dsigl>1e-10));numiter++){
      if(numiter>=maxiter){
	//Failed to converge
	throw Distribution_error("Could not find MLE estimate for log-normal parameters.\n");
      };
      long double dddet=ddmul*ddsigl-ddmusigl*ddmusigl;
      long double ddtr=ddmul+ddsigl;
      //Check for negative definiteness
      if(dddet>0){
	if(ddtr<0){
	  //negative definite. Should be OK
	  mu-=(ddsigl*dmul-ddmusigl*dsigl)/dddet;
	  long double sigstep=(ddmul*dsigl-ddmusigl*dmul)/dddet;
	  if(sigma<sigstep){
	    sigma/=2;
	  }else{
	    sigma-=sigstep;
	  };
	}else{
	  //positive definite.
	  mu=(ddsigl*dmul-ddmusigl*dsigl)/dddet;
	  long double sigstep=(ddmul*dsigl-ddmusigl*dmul)/dddet;
	  if(sigma+sigstep<0){
	    sigma/=2;
	  }else{
	    sigma+=sigstep;
	  };	
	};
      }else{
	//Saddle point
	//use gradient ascent
	mu+=dmul/abs(ddmul);
	long double sigmastep=dsigl/abs(ddsigl);
	if(sigma+sigmastep>0){
	  sigma+=sigmastep;
	}else{
	  sigma/=2;
	};
      };
      this->evalprobsandderivstodirect(maxdata,probsandderivs);
      unsigned int firstval=0;
      for(unsigned int i=0;i<dat.n;i++){
	for(unsigned int j=0;j<6;j++){
	  *(panddgrouped+i*6+j)=0;
	};
	for(unsigned int k=firstval;k<*(data+i);k++){
	  for(unsigned int j=0;j<6;j++){
	    *(panddgrouped+i*6+j)+=*(probsandderivs+k*6+j);
	  };
	};
      };
      
      dmul=0;      
      dsigl=0;	   
      ddmul=0;	   
      ddsigl=0;	   
      ddmusigl=0;
      for(unsigned int i=0;i<=maxdata;i++){
	if(*(freqs+i)){	    
	  xdmul=*(panddgrouped+6*i+1)/(*(panddgrouped+6*i));
	  xdsigl=*(panddgrouped+6*i+2)/(*(panddgrouped+6*i));
	  xddmul=*(panddgrouped+6*i+3)/(*(panddgrouped+6*i))-*(panddgrouped+6*i+1)*(*(panddgrouped+6*i+1))/(*(panddgrouped+6*i)*(*(panddgrouped+6*i)));
	  xddmusigl=*(panddgrouped+6*i+4)/(*(panddgrouped+6*i))-*(panddgrouped+6*i+1)*(*(panddgrouped+6*i+2))/(*(panddgrouped+6*i)*(*(panddgrouped+6*i)));
	  xddsigl=*(panddgrouped+6*i+5)/(*(panddgrouped+6*i))-*(panddgrouped+6*i+2)*(*(panddgrouped+6*i+2))/(*(panddgrouped+6*i)*(*(panddgrouped+6*i)));
	  dmul+=xdmul*(*(freqs+i));      
	  dsigl+=xdsigl*(*(freqs+i));	   	   
	  ddmul+=xddmul*(*(freqs+i));	   	   
	  ddsigl+=xddsigl*(*(freqs+i));	   	   
	  ddmusigl+=xddmusigl*(*(freqs+i));  	       		
	};
      };
    };
    delete[] probsandderivs;
    delete[] panddgrouped;
  }else{
    throw Distribution_error("Parameter estimation for log-normal Poisson distribution not currently implemented for this data type.\n");
  };
};







